@@ -19,7 +19,7 @@ import struct
 import sys
 import zlib
 
-from PyInstaller.building.utils import get_code_object, strip_paths_in_code
+from PyInstaller.building.utils import get_code_object, replace_filename_in_code_object
 from PyInstaller.compat import BYTECODE_MAGIC, is_win, strict_collect_mode
 from PyInstaller.loader.pyimod01_archive import PYZ_ITEM_MODULE, PYZ_ITEM_NSPKG, PYZ_ITEM_PKG
 
@@ -77,16 +77,27 @@ class ZlibArchiveWriter:
         name, src_path, typecode = entry
         assert typecode in {'PYMODULE', 'PYMODULE-1', 'PYMODULE-2'}
 
-        typecode = PYZ_ITEM_MODULE
-        if src_path in ('-', None):
-            # This is a NamespacePackage, modulegraph marks them by using the filename '-'. (But wants to use None,
-            # so check for None, too, to be forward-compatible.)
-            typecode = PYZ_ITEM_NSPKG
+        if src_path in {'-', None}:
+            # PEP-420 namespace package; these do not have code objects, but we still need an entry in PYZ to inform our
+            # run-time module finder/loader of the package's existence. So create a TOC entry for 0-byte data blob,
+            # and write no data.
+            return (name, (PYZ_ITEM_NSPKG, fp.tell(), 0))
+
+        code_object = code_dict[name]
+
+        src_basename, _ = os.path.splitext(os.path.basename(src_path))
+        if src_basename == '__init__':
+            typecode = PYZ_ITEM_PKG
+            co_filename = os.path.join(*name.split('.'), '__init__.py')
         else:
-            src_basename, _ = os.path.splitext(os.path.basename(src_path))
-            if src_basename == '__init__':
-                typecode = PYZ_ITEM_PKG
-        data = marshal.dumps(code_dict[name])
+            typecode = PYZ_ITEM_MODULE
+            co_filename = os.path.join(*name.split('.')) + '.py'
+
+        # Replace co_filename on code object with anonymized version without absolute path to the module.
+        code_object = replace_filename_in_code_object(code_object, co_filename)
+
+        # Serialize
+        data = marshal.dumps(code_object)
 
         # First compress, then encrypt.
         obj = zlib.compress(data, cls._COMPRESSION_LEVEL)
@@ -211,7 +222,17 @@ class CArchiveWriter:
             # by the bootloader. For that, we need to know target optimization level, which is stored in typecode.
             optim_level = {'s': 0, 's1': 1, 's2': 2}[typecode]
             code = get_code_object(dest_name, src_name, optimize=optim_level)
-            code = strip_paths_in_code(code)
+            # Construct new `co_filename` by taking destination name, and replace its suffix with the one from the code
+            # object's co_filename; this should cover all of the following cases:
+            #  - run-time hook script: the source name has a suffix (that is also present in `co_filename` produced by
+            #    `get_code_object`), destination name has no suffix.
+            #  - entry-point script with a suffix: both source name and destination name have the same suffix (and the
+            #    same suffix is also in `co_filename` produced by `get_code_object`)
+            #  - entry-point script without a suffix: neither source name nor destination name have a suffix, but
+            #    `get_code_object` adds a .py suffix to `co_filename` to mitigate potential issues with POSIX
+            #    executables and `traceback` module; we want to preserve this behavior.
+            co_filename = os.path.splitext(dest_name)[0] + os.path.splitext(code.co_filename)[1]
+            code = replace_filename_in_code_object(code, co_filename)
             return self._write_blob(fp, marshal.dumps(code), dest_name, 's', compress=compress)
         elif typecode in ('m', 'M'):
             # Read the PYC file. We do not perform compilation here (in contrast to script files in the above branch),
@@ -221,7 +242,8 @@ class CArchiveWriter:
             assert data[:4] == BYTECODE_MAGIC
             # Skip the PYC header, load the code object.
             code = marshal.loads(data[16:])
-            code = strip_paths_in_code(code)
+            co_filename = dest_name + '.py'  # Use dest name with added .py suffix.
+            code = replace_filename_in_code_object(code, co_filename)
             # These module entries are loaded and executed within the bootloader, which requires only the code
             # object, without the PYC header.
             return self._write_blob(fp, marshal.dumps(code), dest_name, typecode, compress=compress)
@@ -314,9 +336,10 @@ class SplashWriter:
     #
     # typedef struct _splash_data_header
     # {
-    #     char tcl_libname[16];
-    #     char tk_libname[16];
-    #     char tk_lib[16];
+    #     char tcl_shared_library_name[32];
+    #     char tk_shared_library_name[32];
+    #     char tcl_module_directory_name[16];
+    #     char tk_module_directory_name[16];
     #
     #     uint32_t script_len;
     #     uint32_t script_offset;
@@ -328,20 +351,31 @@ class SplashWriter:
     #     uint32_t requirements_offset;
     # } SPLASH_DATA_HEADER;
     #
-    _HEADER_FORMAT = '!32s 32s 16s II II II'
+    _HEADER_FORMAT = '!32s 32s 16s 16s II II II'
     _HEADER_LENGTH = struct.calcsize(_HEADER_FORMAT)
 
     # The created archive is compressed by the CArchive, so no need to compress the data here.
 
-    def __init__(self, filename, name_list, tcl_libname, tk_libname, tklib, image, script):
+    def __init__(
+        self,
+        filename,
+        requirements_list,
+        tcl_shared_library_name,
+        tk_shared_library_name,
+        tcl_module_directory_name,
+        tk_module_directory_name,
+        image,
+        script,
+    ):
         """
         Writer for splash screen resources that are bundled into the CArchive as a single archive/entry.
 
         :param filename: The filename of the archive to create
-        :param name_list: List of filenames for the requirements array
-        :param str tcl_libname: Name of the tcl shared library file
-        :param str tk_libname: Name of the tk shared library file
-        :param str tklib: Root of tk library (e.g. tk/)
+        :param requirements_list: List of filenames for the requirements array
+        :param str tcl_shared_library_name: Basename of the Tcl shared library
+        :param str tk_shared_library_name: Basename of the Tk shared library
+        :param str tcl_module_directory_name: Basename of the Tcl module directory (e.g., tcl/)
+        :param str tk_module_directory_name: Basename of the Tk module directory (e.g., tk/)
         :param Union[str, bytes] image: Image like object
         :param str script: The tcl/tk script to execute to create the screen.
         """
@@ -356,7 +390,7 @@ class SplashWriter:
                 filename = filename.replace(os.path.sep, '\\')
             return filename
 
-        name_list = [_normalize_filename(name) for name in name_list]
+        requirements_list = [_normalize_filename(name) for name in requirements_list]
 
         with open(filename, "wb") as fp:
             # Reserve space for the header.
@@ -367,7 +401,7 @@ class SplashWriter:
             # null-byte, that keeps the list short memory wise and makes it iterable from C.
             requirements_len = 0
             requirements_offset = fp.tell()
-            for name in name_list:
+            for name in requirements_list:
                 name = name.encode('utf-8') + b'\0'
                 fp.write(name)
                 requirements_len += len(name)
@@ -408,9 +442,10 @@ class SplashWriter:
             # Write header
             header_data = struct.pack(
                 self._HEADER_FORMAT,
-                _encode_str(tcl_libname, 'tcl_libname', 32),
-                _encode_str(tk_libname, 'tk_libname', 32),
-                _encode_str(tklib, 'tklib', 16),
+                _encode_str(tcl_shared_library_name, 'tcl_shared_library_name', 32),
+                _encode_str(tk_shared_library_name, 'tk_shared_library_name', 32),
+                _encode_str(tcl_module_directory_name, 'tcl_module_directory_name', 16),
+                _encode_str(tk_module_directory_name, 'tk_module_directory_name', 16),
                 script_len,
                 script_offset,
                 image_len,
