@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 #-----------------------------------------------------------------------------
 # Copyright (c) 2005-2023, PyInstaller Development Team.
 #
@@ -18,7 +17,7 @@ from PyInstaller import isolated
 from PyInstaller.compat import is_win, is_darwin, is_linux
 from PyInstaller.utils.hooks import check_requirement, can_import_module
 from PyInstaller.utils.hooks.qt import get_qt_library_info
-from PyInstaller.utils.tests import importorskip, requires, skipif
+from PyInstaller.utils.tests import importorskip, requires, skipif, onedir_only
 
 
 def qt_param(qt_flavor, *args, **kwargs):
@@ -33,8 +32,8 @@ def qt_param(qt_flavor, *args, **kwargs):
 _QT_PY_PACKAGES = ['PyQt5', 'PyQt6', 'PySide2', 'PySide6']
 QtPyLibs = pytest.mark.parametrize('QtPyLib', [qt_param(i) for i in _QT_PY_PACKAGES])
 
-# OS X bundles, produced by the ``--windowed`` flag, invoke a unique code path that sometimes causes failures in Qt
-# applications.
+# macOS .app bundles, produced by the ``--windowed`` flag, invoke a unique code path that sometimes causes failures in
+# Qt applications. So build with ``--windowed`` option, which will build and run both POSIX build and .app bundle.
 USE_WINDOWED_KWARG = dict(pyi_args=['--windowed']) if is_darwin else {}
 
 
@@ -63,13 +62,13 @@ _ensure_qt_library_info_is_initialized()
 @QtPyLibs
 def test_Qt_QtWidgets(pyi_builder, QtPyLib):
     pyi_builder.test_source(
-        """
+        f"""
         import sys
 
-        from {0}.QtWidgets import QApplication, QWidget
-        from {0}.QtCore import QTimer
+        from {QtPyLib}.QtWidgets import QApplication, QWidget
+        from {QtPyLib}.QtCore import QTimer
 
-        is_qt6 = '{0}' in {{'PySide6', 'PyQt6'}}
+        is_qt6 = '{QtPyLib}' in {{'PySide6', 'PyQt6'}}
 
         app = QApplication(sys.argv)
         window = QWidget()
@@ -85,7 +84,7 @@ def test_Qt_QtWidgets(pyi_builder, QtPyLib):
         else:
             res = app.exec_()
         sys.exit(res)
-        """.format(QtPyLib), **USE_WINDOWED_KWARG
+        """, **USE_WINDOWED_KWARG
     )
 
 
@@ -103,14 +102,14 @@ def test_Qt_QtQml(pyi_builder, QtPyLib):
             pytest.skip('PySide6-Essentials 6.6.3 is missing shared libraries required by Qt Quick Controls 2.')
 
     pyi_builder.test_source(
-        """
+        f"""
         import sys
 
-        from {0}.QtGui import QGuiApplication
-        from {0}.QtQml import QQmlApplicationEngine
-        from {0}.QtCore import QTimer, QUrl
+        from {QtPyLib}.QtGui import QGuiApplication
+        from {QtPyLib}.QtQml import QQmlApplicationEngine
+        from {QtPyLib}.QtCore import QTimer, QUrl
 
-        is_qt6 = '{0}' in {{'PyQt6', 'PySide6'}}
+        is_qt6 = '{QtPyLib}' in {{'PyQt6', 'PySide6'}}
 
         # Select a style via the `command line
         # <https://doc.qt.io/qt-5/qtquickcontrols2-styles.html#command-line-argument>`_,
@@ -144,7 +143,7 @@ def test_Qt_QtQml(pyi_builder, QtPyLib):
             res = app.exec_()
         del engine
         sys.exit(res)
-        """.format(QtPyLib), **USE_WINDOWED_KWARG
+        """, **USE_WINDOWED_KWARG
     )
 
 
@@ -182,10 +181,10 @@ def test_Qt_QtNetwork_SSL_support(pyi_builder, QtPyLib):
         pytest.skip('QtNetwork does not use OpenSSL.')
 
     pyi_builder.test_source(
-        """
+        f"""
         import sys
-        from {0}.QtCore import QCoreApplication, QLibraryInfo
-        from {0}.QtNetwork import QSslSocket
+        from {QtPyLib}.QtCore import QCoreApplication, QLibraryInfo
+        from {QtPyLib}.QtNetwork import QSslSocket
 
         app = QCoreApplication(sys.argv)
 
@@ -213,17 +212,17 @@ def test_Qt_QtNetwork_SSL_support(pyi_builder, QtPyLib):
             print(f"Active TLS backend: {{QSslSocket.activeBackend()}}")
             print(f"Available TLS backends: {{QSslSocket.availableBackends()}}")
             assert 'openssl' in QSslSocket.availableBackends(), "OpenSSL TLS backend not available!"
-        """.format(QtPyLib), **USE_WINDOWED_KWARG
+        """, **USE_WINDOWED_KWARG
     )
 
 
 @QtPyLibs
 def test_Qt_QTranslate(pyi_builder, QtPyLib):
     pyi_builder.test_source(
-        """
+        f"""
         import sys
-        from {0}.QtWidgets import QApplication
-        from {0}.QtCore import QTranslator, QLocale, QLibraryInfo
+        from {QtPyLib}.QtWidgets import QApplication
+        from {QtPyLib}.QtCore import QTranslator, QLocale, QLibraryInfo
 
         # Initialize Qt default translations
         app = QApplication(sys.argv)
@@ -243,27 +242,24 @@ def test_Qt_QTranslate(pyi_builder, QtPyLib):
         else:
             print('Qt locale %s not found!' % locale.name())
             assert False
-        """.format(QtPyLib)
+        """
     )
 
 
 @QtPyLibs
-def test_Qt_Ui_file(tmpdir, pyi_builder, data_dir, QtPyLib):
-    # Note that including the data_dir fixture copies files needed by this test.
+def test_Qt_Ui_file(pyi_builder, data_dir, QtPyLib):
     pyi_builder.test_source(
-        """
+        f"""
         import os
         import sys
 
-        import {0}.QtQuickWidgets  # Used instead of hiddenimports
+        import {QtPyLib}.QtQuickWidgets  # Used instead of hiddenimports
 
-        from {0}.QtWidgets import QApplication, QWidget
-        from {0}.QtCore import QTimer
+        from {QtPyLib}.QtWidgets import QApplication, QWidget
+        from {QtPyLib}.QtCore import QTimer
 
-        from pyi_get_datadir import get_data_dir
-
-        is_qt6 = '{0}' in {{'PyQt6', 'PySide6'}}
-        is_pyqt = '{0}' in {{'PyQt5', 'PyQt6'}}
+        is_qt6 = '{QtPyLib}' in {{'PyQt6', 'PySide6'}}
+        is_pyqt = '{QtPyLib}' in {{'PyQt5', 'PyQt6'}}
 
         app = QApplication(sys.argv)
 
@@ -273,21 +269,21 @@ def test_Qt_Ui_file(tmpdir, pyi_builder, data_dir, QtPyLib):
         if is_qt6:
             try:
                 # This seems to be unsupported on macOS version of PySide6 at the time of writing (6.1.0)
-                from {0}.QtQuick import QQuickWindow, QSGRendererInterface
+                from {QtPyLib}.QtQuick import QQuickWindow, QSGRendererInterface
                 QQuickWindow.setGraphicsApi(QSGRendererInterface.GraphicsApi.OpenGL)
             except Exception:
                 pass
 
         # Load the UI
-        ui_file = os.path.join(get_data_dir(), 'Qt_Ui_file', 'gui.ui')
+        ui_file = os.path.join(os.path.dirname(__file__), 'gui.ui')
         if is_pyqt:
             # Use PyQt.uic
-            from {0} import uic
+            from {QtPyLib} import uic
             window = QWidget()
             uic.loadUi(ui_file, window)
         else:
             # Use PySide.QtUiTools.QUiLoader
-            from {0}.QtUiTools import QUiLoader
+            from {QtPyLib}.QtUiTools import QUiLoader
             loader = QUiLoader()
             window = loader.load(ui_file)
         window.show()
@@ -302,7 +298,9 @@ def test_Qt_Ui_file(tmpdir, pyi_builder, data_dir, QtPyLib):
         else:
             res = app.exec_()
         sys.exit(res)
-        """.format(QtPyLib)
+        """,
+        # Collect the .ui file into top-level application directory.
+        pyi_args=['--add-data', f"{data_dir / 'gui.ui'}:."],
     )
 
 
@@ -321,7 +319,7 @@ def test_Qt_Ui_file(tmpdir, pyi_builder, data_dir, QtPyLib):
 @requires('PyQt5')
 @pytest.mark.skipif(
     check_requirement('PyQt5 == 5.11.3') and is_darwin,
-    reason='This version of the OS X wheel does not include QWebEngine.'
+    reason='This version of the macOS wheel does not include QWebEngine.'
 )
 def test_PyQt5_Qt(pyi_builder):
     pyi_builder.test_source('from PyQt5.Qt import QLibraryInfo', **USE_WINDOWED_KWARG)
@@ -364,19 +362,21 @@ def _disable_qtwebengine_sandbox(qt_flavor):
 
 # Run the the QtWebEngineWidgets test for chosen Qt-based package flavor.
 def _test_Qt_QtWebEngineWidgets(pyi_builder, qt_flavor):
-    source = """
+    disable_sandbox = _disable_qtwebengine_sandbox(qt_flavor)
+    pyi_builder.test_source(
+        f"""
         import sys
 
         # Disable QtWebEngine/chromium sanbox, if necessary
-        if {1}:
+        if {disable_sandbox}:
             import os
             os.environ['QTWEBENGINE_DISABLE_SANDBOX'] = '1'
 
-        from {0}.QtWidgets import QApplication
-        from {0}.QtWebEngineWidgets import QWebEngineView
-        from {0}.QtCore import QTimer
+        from {qt_flavor}.QtWidgets import QApplication
+        from {qt_flavor}.QtWebEngineWidgets import QWebEngineView
+        from {qt_flavor}.QtCore import QTimer
 
-        is_qt6 = '{0}' in {{'PyQt6', 'PySide6'}}
+        is_qt6 = '{qt_flavor}' in {{'PyQt6', 'PySide6'}}
 
         # Web page to display
         WEB_PAGE_HTML = '''
@@ -417,10 +417,12 @@ def _test_Qt_QtWebEngineWidgets(pyi_builder, qt_flavor):
 
             def verify_and_quit(self):
                 # Make sure the renderer process is alive.
+                print("Checking the result of renderer process...", file=sys.stderr)
                 if self.result != self.EXPECTED:
                     raise ValueError(
                         f"JS result is {{self.result!r}} but expected {{self.EXPECTED!r}}. "
                         "Is the QtWebEngine renderer process running properly?")
+                print("Exiting application's main loop...", file=sys.stderr)
                 app.quit()
 
         view = QWebEngineView()
@@ -430,36 +432,43 @@ def _test_Qt_QtWebEngineWidgets(pyi_builder, qt_flavor):
         js_result_tester = JSResultTester()
         js_result_tester.setup(view)
 
+        print("Entering application's main loop...", file=sys.stderr)
         if is_qt6:
             # Qt6: exec_() is deprecated in PySide6 and removed from PyQt6 in favor of exec()
             res = app.exec()
         else:
             res = app.exec_()
-        sys.exit(res)
-        """.format(qt_flavor, _disable_qtwebengine_sandbox(qt_flavor))
+        print("Exited application's main loop!", file=sys.stderr)
 
-    pyi_builder.test_source(source, **USE_WINDOWED_KWARG)
+        print("Calling sys.exit()...", file=sys.stderr)
+        sys.exit(res)
+        """, **USE_WINDOWED_KWARG
+    )
 
 
 # Run the the QtWebEngineQuick test for chosen Qt-based package flavor.
 def _test_Qt_QtWebEngineQuick(pyi_builder, qt_flavor):
-    source = """
+    disable_sandbox = _disable_qtwebengine_sandbox(qt_flavor)
+    pyi_builder.test_source(
+        f"""
         import sys
 
         # Disable QtWebEngine/chromium sanbox, if necessary
-        if {1}:
+        if {disable_sandbox}:
             import os
             os.environ['QTWEBENGINE_DISABLE_SANDBOX'] = '1'
 
-        from {0}.QtGui import QGuiApplication
-        from {0}.QtQml import QQmlApplicationEngine
+        from {qt_flavor}.QtGui import QGuiApplication
+        from {qt_flavor}.QtQml import QQmlApplicationEngine
 
-        is_qt6 = '{0}' in {{'PyQt6', 'PySide6'}}
+        is_qt6 = '{qt_flavor}' in {{'PyQt6', 'PySide6'}}
 
         if is_qt6:
-            from {0}.QtWebEngineQuick import QtWebEngineQuick
+            from {qt_flavor}.QtWebEngineQuick import QtWebEngineQuick
         else:
-            from {0}.QtWebEngine import QtWebEngine as QtWebEngineQuick
+            from {qt_flavor}.QtWebEngine import QtWebEngine as QtWebEngineQuick
+
+        # Must be called before QGuiApplication is instantiated!
         QtWebEngineQuick.initialize()
 
         app = QGuiApplication(sys.argv)
@@ -474,43 +483,59 @@ def _test_Qt_QtWebEngineQuick(pyi_builder, qt_flavor):
                 WebEngineView {{
                     id: view
                     anchors.fill: parent
-                    Component.onCompleted: loadHtml('
-                        <!doctype html>
-                        <html lang="en">
-                            <head>
-                                <meta charset="utf-8">
-                                <title>Test web page</title>
-                            </head>
-                            <body>
-                                <p>This is a test web page.</p>
-                            </body>
-                        </html>
-                    ')
-                }}
-                Connections {{
-                    target: view
-                    function onLoadingChanged(loadRequest) {{
+                    Component.onCompleted: () => {{
+                        console.info("Loading HTML...")
+                        loadHtml('
+                            <!doctype html>
+                            <html lang="en">
+                                <head>
+                                    <meta charset="utf-8">
+                                    <title>Test web page</title>
+                                </head>
+                                <body>
+                                    <p>This is a test web page.</p>
+                                </body>
+                            </html>
+                        ')
+                    }}
+                    onLoadingChanged: (loadRequest) => {{
+                        console.info("Web page loading status changed: " + loadRequest.status)
                         if (loadRequest.status !== WebEngineView.LoadStartedStatus) {{
-                            Qt.quit()
+                            console.info("Page loading finished; running shutdown timer (" + timer.interval + " ms)!")
+                            timer.running = true
                         }}
                     }}
                 }}
+                Timer {{
+                    id: timer
+                    interval: 1000
+                    running: false
+                    repeat: false
+                    onTriggered: () => {{
+                        console.info("Shutdown timer triggered; exiting application's main loop...")
+                        Qt.quit()
+                    }}
+                }}
             }}
         ''')
 
         if not engine.rootObjects():
-            sys.exit(-1)
+            raise RuntimeError("No root objects loaded from QML!")
 
+        print("Entering application's main loop...", file=sys.stderr)
         if is_qt6:
             # Qt6: exec_() is deprecated in PySide6 and removed from PyQt6 in favor of exec()
             res = app.exec()
         else:
             res = app.exec_()
-        del engine
-        sys.exit(res)
-        """.format(qt_flavor, _disable_qtwebengine_sandbox(qt_flavor))
+        print("Exited application's main loop!", file=sys.stderr)
 
-    pyi_builder.test_source(source, **USE_WINDOWED_KWARG)
+        del engine
+
+        print("Calling sys.exit()...", file=sys.stderr)
+        sys.exit(res)
+        """, **USE_WINDOWED_KWARG
+    )
 
 
 @requires('PyQt5')
@@ -537,6 +562,11 @@ def test_Qt_QtWebEngineQuick_PySide2(pyi_builder):
 
 @requires('PyQt6 >= 6.2.2')
 @requires('PyQt6-WebEngine')  # NOTE: base Qt6 must be 6.2.2 or newer, QtWebEngine can be older
+@pytest.mark.flaky(
+    # Attempt to mitigate issues with QtWebEngine 6.10.1
+    condition=check_requirement('PyQt6-WebEngine-Qt6 == 6.10.1'),
+    reruns=1,
+)
 def test_Qt_QtWebEngineWidgets_PyQt6(pyi_builder):
     _test_Qt_QtWebEngineWidgets(pyi_builder, 'PyQt6')
 
@@ -551,6 +581,12 @@ def test_Qt_QtWebEngineWidgets_PyQt6(pyi_builder):
     check_requirement('PyQt6-Qt6 == 6.6.3') and is_win,
     reason='PyQt6 6.6.3 PyPI wheels for Windows are missing Qt6WebChannelQuick shared library.'
 )
+@pytest.mark.flaky(
+    # The generated .app bundle seems to sporadically freeze during shutdown on GHA macos-14 runners.
+    # Attempt to mitigate issues with QtWebEngine 6.10.1
+    condition=is_darwin or check_requirement('PyQt6-WebEngine-Qt6 == 6.10.1'),
+    reruns=1,
+)
 def test_Qt_QtWebEngineQuick_PyQt6(pyi_builder):
     _test_Qt_QtWebEngineQuick(pyi_builder, 'PyQt6')
 
@@ -560,6 +596,10 @@ def test_Qt_QtWebEngineQuick_PyQt6(pyi_builder):
     check_requirement('PySide6 == 6.5.0') and is_win,
     reason='PySide6 6.5.0 PyPI wheels for Windows are missing opengl32sw.dll.'
 )
+@pytest.mark.flaky(
+    # Attempt to mitigate issues with QtWebEngine 6.10.1
+    condition=check_requirement('PySide6 == 6.10.1'),
+)
 def test_Qt_QtWebEngineWidgets_PySide6(pyi_builder):
     _test_Qt_QtWebEngineWidgets(pyi_builder, 'PySide6')
 
@@ -569,6 +609,12 @@ def test_Qt_QtWebEngineWidgets_PySide6(pyi_builder):
     check_requirement('PySide6 == 6.5.0') and is_win,
     reason='PySide6 6.5.0 PyPI wheels for Windows are missing opengl32sw.dll.'
 )
+@pytest.mark.flaky(
+    # The generated .app bundle seems to sporadically freeze during shutdown on GHA macos-14 runners.
+    # Attempt to mitigate issues with QtWebEngine 6.10.1
+    condition=is_darwin or check_requirement('PySide6 == 6.10.1'),
+    reruns=1,
+)
 def test_Qt_QtWebEngineQuick_PySide6(pyi_builder):
     _test_Qt_QtWebEngineQuick(pyi_builder, 'PySide6')
 
@@ -583,14 +629,14 @@ def test_Qt_QtMultimedia_player_init(pyi_builder, QtPyLib):
             pytest.skip('QtMultimedia is broken under PyQt6 6.7.0 and PyQt6-Qt6 6.7.1.')
 
     pyi_builder.test_source(
-        """
+        f"""
         import sys
 
-        from {0} import QtCore, QtMultimedia
+        from {QtPyLib} import QtCore, QtMultimedia
 
         app = QtCore.QCoreApplication(sys.argv)
         player = QtMultimedia.QMediaPlayer(app)
-        """.format(QtPyLib), **USE_WINDOWED_KWARG
+        """, **USE_WINDOWED_KWARG
     )
 
 
@@ -606,13 +652,13 @@ def test_Qt_QtMultimedia_player_init(pyi_builder, QtPyLib):
 ])
 def test_Qt_QtMultimedia_with_true_property(pyi_builder, QtPyLib):
     pyi_builder.test_source(
-        """
+        f"""
         import sys
-        from {0} import QtCore, QtMultimedia
+        from {QtPyLib} import QtCore, QtMultimedia
         from __feature__ import true_property
 
         app = QtCore.QCoreApplication(sys.argv)
-        """.format(QtPyLib), **USE_WINDOWED_KWARG
+        """, **USE_WINDOWED_KWARG
     )
 
 
@@ -671,27 +717,27 @@ def _test_qt_bindings_import(bindings, module, pyi_builder_onedir):
 
 @importorskip('PySide2')
 @pytest.mark.parametrize('module', _list_all_qt_submodules('PySide2'))
-@pytest.mark.parametrize('pyi_builder', ['onedir'], indirect=True)
+@onedir_only
 def test_qt_module_import_PySide2(module, pyi_builder):
     _test_qt_bindings_import("PySide2", module, pyi_builder)
 
 
 @importorskip('PySide6')
 @pytest.mark.parametrize('module', _list_all_qt_submodules('PySide6'))
-@pytest.mark.parametrize('pyi_builder', ['onedir'], indirect=True)
+@onedir_only
 def test_qt_module_import_PySide6(module, pyi_builder):
     _test_qt_bindings_import("PySide6", module, pyi_builder)
 
 
 @importorskip('PyQt5')
 @pytest.mark.parametrize('module', _list_all_qt_submodules('PyQt5'))
-@pytest.mark.parametrize('pyi_builder', ['onedir'], indirect=True)
+@onedir_only
 def test_qt_module_import_PyQt5(module, pyi_builder):
     _test_qt_bindings_import("PyQt5", module, pyi_builder)
 
 
 @importorskip('PyQt6')
 @pytest.mark.parametrize('module', _list_all_qt_submodules('PyQt6'))
-@pytest.mark.parametrize('pyi_builder', ['onedir'], indirect=True)
+@onedir_only
 def test_qt_module_import_PyQt6(module, pyi_builder):
     _test_qt_bindings_import("PyQt6", module, pyi_builder)
